
/*  @(#)cback.c 1.5 93/06/01
 *
 *  A color random background generator.
 *
 *  Color squares inspired by James Goslings' Mondrian picture generator.
 *  Color circles based on a program by Nigel Bond on a Perq in 1983.
 *  Wallpaper patterns based on Computer Recreations article by A. K. Dewdney
 *  in Scientific American.
 *
 *  Copyright (C) 1989 - 1993 - Rich Burridge - Sun Microsystems.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged.
 *
 *  No responsibility is taken for any errors or inaccuracies inherent
 *  either to the comments or the code of this program, but if
 *  reported to me then an attempt will be made to fix them.
 *
 *  CHANGE LOG:
 *
 *  v1.3 [2nd June 1993] - replaced <time.h> with <sys/time.h>
 *                       - now look for ConfigureNotify events.
 *                       - added wallpaper option.
 *
 *  v1.2 [18th May 1993] - released to alt.sources.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <poll.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define  PATCHLEVEL   3

/* For all function declarations, if ANSI then use a prototype. */
 
#if  defined(__STDC__)
#define P(args)  args
#else  /* ! __STDC__ */
#define P(args)  ()
#endif  /* STDC */

#define  FPRINTF      (void) fprintf
#define  FREE         (void) free
#define  POLL         (void) poll
#define  STRCPY       (void) strcpy

#define  CB_SIZE      32      /* Cback colormap size. */
#define  INC          argc-- , argv++
#define  MAXLINE      256     /* Maximum length of character strings. */
#define  SECSINDAY    86400   /* Number of seconds in a day. */

#define  WHITE         0 
#define  BLACK         CB_SIZE-1

int cmap[CB_SIZE] ;       /* Current cback colormap. */
int color ;               /* Current color of square or circle. */
int csize ;               /* Colormap size. */
int dir ;                 /* Current direction of square chain. */
int len ;                 /* Current length of square chain. */
int maxno ;               /* Maximum number of circles or squares. */
int norows ;              /* Number of rows of squares. */
int nocols ;              /* Number of columns of squares. */
int period ;              /* Time in seconds between random draws. */
int pval ;                /* Current random start position of square chain. */
int screen ;
int ssize ;               /* Square size. */
int total ;               /* Current number of displayed squares. */
int win_depth ;           /* Depth of the screen in pixels. */
int win_height ;          /* Height of the cback canvas in pixels. */
int win_width ;           /* Width of the cback canvas in pixels. */

char option ;                /* Selection random option. */
char progname[MAXLINE] ;     /* This programs name. */
char *sqs ;                  /* Pointer to square array. */
char x11_display[MAXLINE] ;  /* X11 display information. */

struct pollfd pfd[1] ;

unsigned char red[CB_SIZE], green[CB_SIZE], blue[CB_SIZE] ;

Colormap colormap ;
Display *dpy ;
GC gc ;
Pixmap pr ;
Window root ;
XEvent event ;
XGCValues gc_val ;

int get_rdir        P(()) ;
int get_start       P(()) ;
int init_ws_type    P(()) ;
int main            P((int, char **)) ;
int rrange          P((int, int)) ;

void change_cmap    P((unsigned char *, unsigned char *,
                       unsigned char *, int)) ;
void color_area     P((int, int, int, int, int)) ;
void do_redraw      P(()) ;
void draw_circle    P((int, int, int, int)) ;
void draw_square    P((int)) ;
void fill_screen    P(()) ;
void get_options    P((int, char **)) ;
void getparam       P((char *, char **, char *)) ;
void init_colors    P((unsigned char *, unsigned char *, unsigned char *)) ;
void initialise     P(()) ;
void load_colors    P((unsigned char *, unsigned char *,
                       unsigned char *, int)) ;
void make_array     P((int)) ;
void make_circles   P(()) ;
void make_paper     P(()) ;
void make_squares   P(()) ;
void redraw_circle  P(()) ;
void redraw_square  P(()) ;
void start_tool     P((int)) ;


/*ARGSUSED*/
int
main(argc, argv)
int argc ;
char *argv[] ;
{
  STRCPY(progname, argv[0]) ;       /* Save for later use. */
  get_options(argc, argv) ;         /* Extract command line options. */
  initialise() ;                    /* Initialise variables, arrays etc.. */
  if (init_ws_type())               /* Determine window system type. */
    {
      FPRINTF(stderr, "Error initialising window system.\n") ;
      exit(1) ;
    }
  init_colors(red, green, blue) ;          /* Initialise colormap entries. */
  load_colors(red, green, blue, csize) ;   /* Load cback colormap. */
  fill_screen() ;                          /* Initial random creation. */
  do_redraw() ;
  start_tool(period) ;
  exit(0) ;
/*NOTREACHED*/
}


void
change_cmap(red, green, blue, size)   /* Change colormap entries. */
unsigned char *red, *green, *blue ;
int size ;
{
  XColor curcolor ;
  int i ;
  static int firsttime = 1 ;
 
  if (!firsttime) XFreeColors(dpy, colormap, (unsigned long *) cmap, size, 0) ;
  for (i = 0; i < size; i++)
    {
      curcolor.flags = DoRed | DoGreen | DoBlue ;
      curcolor.red = (unsigned short) (red[i] << 8) ;
      curcolor.green = (unsigned short) (green[i] << 8) ;
      curcolor.blue = (unsigned short) (blue[i] << 8) ;
      if (XAllocColor(dpy, colormap, &curcolor) == True)
        cmap[i] = curcolor.pixel ;
    }
  firsttime = 0 ;
}


void
color_area(x, y, width, height, color)
int x, y, width, height, color ;
{
  XSetForeground(dpy, gc, (unsigned long) cmap[color]) ;
  XFillRectangle(dpy, pr, gc, x, y,
                 (unsigned int) width, (unsigned int) height) ;
}


void
do_redraw()      /* Draw new circles or squares onto the screen. */
{
  XCopyArea(dpy, pr, root, gc, 0, 0,
            (unsigned int) win_width, (unsigned int) win_height, 0, 0) ;
}


void
draw_circle(cx, cy, r, color)     /* Draw a circle in a given color. */
int cx, cy, r, color ;
{
  if (r <= 0) return ;
  XSetForeground(dpy, gc, (unsigned long) cmap[color]) ;
  XFillArc(dpy, pr, gc, cx - r, cy - r, 2 * r, 2 * r, 0, 23040) ;
}


void
draw_square(val)     /* Draw colored square at position pos. */
int val ;
{
  color_area((val % nocols) * ssize,        /* White background first. */
             (int) ((val / nocols) * ssize), ssize, ssize, WHITE) ;
  color_area((val % nocols) * ssize + 2,    /* Colored square inside. */
             (int) ((val / nocols) * ssize + 2), ssize-4, ssize-4, color) ;
  sqs[val] = color ;
  pval = val ;
  total++ ;
}


void
fill_screen()       /* Draw next screen load of squares or circles. */
{
  init_colors(red, green, blue) ;   /* Initialise colormap entries. */
  color_area(0, 0, win_width, win_height, BLACK) ;  /* Blackout the window. */
  switch (option)
    {
      case 'c' : make_circles() ;   /* Generate random circles. */
                 break ;
      case 's' : make_squares() ;   /* Generate random squares. */
                 break ;
      case 'w' : make_paper() ;     /* Generate wallpaper patterns. */
    }
}


void
get_options(argc, argv)      /* Extract command line options. */
int argc ;
char *argv[] ;
{
  char next[MAXLINE] ;       /* The next command line parameter. */

  period = 300 ;
  option = 's' ;
  INC ;
  while (argc > 0)
    {
      if (argv[0][0] == '-')
        switch (argv[0][1])
          {
            case 'c' : option = 'c' ;     /* Generate random circles. */
                       break ;
            case 'd' : INC ;              /* X11 display information. */
                       getparam(x11_display, argv,
                                "-d needs display information") ;
                       break ;
            case 'p' : INC ;
                       getparam(next, argv, "-p needs period duration") ;
                       period = atoi(next) ;

/*  Check the user hasn't given an invalid value, a negative value, a
 *  zero value or a period greater than one day. If so, set it back to
 *  the default.
 */
                       if (period < 0 || period > SECSINDAY)
                         period = 300 ;
                       break ;
            case 's' : option = 's' ;     /* Generate random squares. */
                       break ;
            case 'v' : FPRINTF(stderr, "%s version 1.0.%1d\n", progname,
                                        PATCHLEVEL) ;
                       break ;
            case 'w' : option = 'w' ;     /* Wallpaper patterns. */
                       break ;
            default  : FPRINTF(stderr, "Usage: %s: [-c] ", progname) ;
                       FPRINTF(stderr, "[-d display] [-p period] [-s] ") ;
                       FPRINTF(stderr, "[-v] [-w]\n") ;
                       exit(1) ;
          }
      INC ;
    }
}


void
getparam(s, argv, errmes)
char *s, *argv[], *errmes ;
{
  if (*argv != NULL && argv[0][0] != '-') STRCPY(s, *argv) ;
  else
    { 
      FPRINTF(stderr, "%s: %s as next argument.\n", progname, errmes) ;
      exit(1) ;
    }
}


int
get_rdir()        /* Get random direction. */
{
  int d ;

  d = rrange(0, 3) ;
  switch (d)
    {
      case 0 : d = ((pval - nocols) < 0) ? 0 : -nocols ;       /* North. */
               break ;
      case 1 : d = ((pval - 1) < 0) ? 0 : -1 ;                 /* West. */
               break ;
      case 2 : d = ((pval + 1) >= maxno) ? 0 : 1 ;             /* East. */
               break ;
      case 3 : d = ((pval + nocols) >= maxno) ? 0 : nocols ;   /* South. */
    }
  return(d) ;
}


int
get_start()       /* Get random start position. */
{
  int p ;

  p = rrange(0, maxno-1) ;
  while (sqs[p] != '\0') p = (p + 1) % maxno ;
  return(p) ;
}


void
init_colors(red, green, blue)   /* Initialise colormap entries. */
unsigned char *red, *green, *blue ;
{
  int good_color, i, inc, ired, igreen, iblue, rinc, ginc, binc ;

  inc = rrange(1, 6) ;          /* Colormap increment. */
  do
    {
      rinc = rrange(0, 1) ;     /* If set, this color is incremented. */
      ginc = rrange(0, 1) ;
      binc = rrange(0, 1) ;
    }
  while (!rinc && !ginc && !binc) ;  /* At least one must increment. */
  ired   = rrange(0, 255) ;     /* Random start positions in the colormap. */
  igreen = rrange(0, 255) ;
  iblue  = rrange(0, 255) ;

  red[WHITE] = green[WHITE] = blue[WHITE] = 255 ;  /* First value is white. */
  red[BLACK] = green[BLACK] = blue[BLACK] = 0 ;    /* Last value is black. */

  for (i = 1; i < csize-1; i++)
    {
      do
        {
          good_color = 1 ;
          red[i]   = (ired   + (inc * i * rinc)) % 256 ;
          green[i] = (igreen + (inc * i * ginc)) % 256 ;
          blue[i]  = (iblue  + (inc * i * binc)) % 256 ;
          if (red[i] > 245 && green[i] > 245 && blue[i] > 245)
            good_color = 0 ;
          if (red[i] < 10  && green[i] < 10  && blue[i] < 10)
            good_color = 0 ;
        }
      while (!good_color) ;
    }
}


void
initialise()                 /* Initialise variables, arrays etc.. */
{
  srand((int) time((time_t *) 0)) ;
  csize = CB_SIZE ;          /* Colormap size. */
  sqs = NULL ;               /* No square array initially. */
}


int
init_ws_type()
{
  if ((dpy = XOpenDisplay(x11_display)) == NULL)
    {
      FPRINTF(stderr,"%s: Couldn't open display %s\n", progname,
              (getenv("DISPLAY") ? getenv("DISPLAY") : x11_display)) ;
      exit(1) ;
    }
 
  pfd[0].fd     = ConnectionNumber(dpy) ;
  pfd[0].events = POLLIN ;

  root       = XDefaultRootWindow(dpy) ;
  screen     = DefaultScreen(dpy) ;
  win_height = DisplayHeight(dpy, screen) ;
  win_width  = DisplayWidth(dpy, screen) ;
  win_depth  = DefaultDepth(dpy, screen) ;
  gc_val.foreground = gc_val.background = cmap[0] ;
  gc = XCreateGC(dpy, root, GCForeground | GCBackground , &gc_val) ;
  pr = XCreatePixmap(dpy, root, (unsigned int) win_width,
                                (unsigned int) win_height, win_depth) ;
  XClearWindow(dpy, root) ;
  XFlush(dpy) ;
  return(0) ;
}


void
load_colors(red, green, blue, size)           /* Load cback color map. */
unsigned char *red, *green, *blue ;
int size ;
{
  colormap = XDefaultColormap(dpy, screen) ;
  if (DisplayCells(dpy, screen) > 2)
    change_cmap(red, green, blue, size) ;     /* Change colormap entries. */
}


void
make_array(ssize)                   /* Create array for square testing. */
int ssize ;
{
  if (sqs != NULL) free(sqs) ;
  norows = win_height / ssize + 1 ;
  nocols = win_width / ssize + 1 ;
  maxno = norows * nocols ;
  sqs = malloc((unsigned) maxno) ;
  memset(sqs, 0, maxno) ;               /* Zeroise character array. */
}


void
make_circles()                      /* Generate random circles. */
{
  int cx, cy, i, r, width ;

  maxno = rrange(50, 200) ;         /* Maximum number of circles. */
  for (total = 0; total < maxno; total++)
    {
      cx    = rrange(0, win_width-1) ;        /* Center X coordinate. */
      cy    = rrange(0, win_height-1) ;       /* Centre Y coordinate. */
      r     = rrange(30, 125) ;               /* Circle radius. */
      color = rrange(1, csize-2) ;            /* Get random color. */
      width = rrange(2, 12) ;                 /* Get random thickness. */
      draw_circle(cx, cy, r+width, color) ;   /* Draw colored circle. */
      draw_circle(cx, cy, r, BLACK) ;         /* Fill with black. */
    }
}


void
make_paper()                        /* Generate wallpaper patterns. */
{
  int base, corna, cornb, i, j, n, side, x, y, z ;

  corna = rrange(-50, 50) ;
  cornb = rrange(-50, 50) ;
  side  = rrange(5, 100) ;
  base  = rrange(100, 250) ;
  for (i = 0; i < win_width; i += 2)
    for (j = 0; j < win_height; j += 2)
      {
        x = corna + (side * i / base) ;
        y = cornb + (side * j / base) ;
        z = (x * x) + (y * y) ;
        n = z % csize ;
        if (n) color_area(i, j, 2, 2, n) ;
      }
}


void
make_squares()                      /* Generate random squares. */
{
  int i ;

  ssize = rrange(60, 250) ;         /* Size of color circle or square. */
  make_array(ssize) ;               /* Create array for square testing. */
  total = 0 ;
  while (total < maxno)             /* While there are squares to be filled. */
    {
      pval = get_start() ;          /* Get random start position. */
      color = rrange(1, csize-2) ;  /* Get random color. */
      draw_square(pval) ;           /* Draw colored square. */
      len = rrange(3, 9) ;          /* Get random length. */
      for (i = 0 ; i < len; i++)
        {
          if ((dir = get_rdir()) == 0)      /* Get random direction. */
            continue ;
          if (sqs[pval + dir] == '\0')      /* Square[pval + dir] empty? */
            draw_square(pval + dir) ;       /* Draw next square in chain. */
          else break ;                      /* Else get new start position. */
        }
    }
}


int
rrange(low, high)   /* Return a random integer between low and high. */
int low, high ;
{
  return((((int) rand() & 077777) % (high - low + 1)) + low) ;
}


void
start_tool(period)    /* Map window and loop of random screenfills. */
int period ;
{
  XEvent event ;                 /* For events received. */

  XSelectInput(dpy, root, SubstructureNotifyMask) ;

  for (;;)
    {
      if (!XPending(dpy)) POLL(pfd, 1L, period * 1000) ;
      if (!XPending(dpy))
        {
          fill_screen() ;                         /* Next random creation. */
          change_cmap(red, green, blue, csize) ;  /* Load next colormap. */
          do_redraw() ;                           /* Load new image. */
        }
      else
        {
          if (!XCheckMaskEvent(dpy, SubstructureNotifyMask, &event))
            XNextEvent(dpy, &event) ;
          if (event.type == MapNotify || event.type == ConfigureNotify)
            do_redraw() ;
        }
    }
}
